// ============================================================================
//
// Unit:    ibeDALConnection
// Inhalt:  Formular
// Aufgabe: DAL Connection Basisklasse
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeDALConnection;

interface

uses
  SysUtils, Classes, Variants, DB,
  ibeTypes, ibeFuncCommon, ibeException, ibeDALQuery;

type

/// <summary> 
/// TibeDALConnection stellt die Datenbankverbindung her. Mit den DAO...-Methoden 
/// kann die Anwendung Query-Objekte instanzieren und konfektionieren mit denen 
/// die Abfragen ausgefhrt werden. 
/// Die CVT..-Methoden wandeln Delphi-Datentypen in die zum DBMS passende String-Notation 
/// um, so dass diese in SQL-Text eingebaut werden knnen. 
/// </summary> 
TibeDALConnection = class
private
  FTranStartCount: Integer;
  FParameters: TStringList;
  FDBMSName, FUserName, FPassword: string;
  FKeepConnection, FTraceEnabled: Boolean;
  procedure SetKeepConnection(AKeep: Boolean);
  function GetTranActive(): Boolean; 
protected
  procedure InternalCreate(); virtual;
  procedure DoAfterInternalCreate(); virtual;
  procedure InternalSetup(); virtual; abstract;
  // Transaktionen steuern 
  procedure InternalTranStart(); virtual; abstract;
  procedure InternalTranCommit(); virtual; abstract;
  procedure InternalTranRollback(aOrigException: Exception = nil); virtual; abstract;
  procedure InternalDAOQueryInit(a_daoQuery: TibeDALQuery); virtual; abstract;
  function InternalDatabaseExists(const ADatabaseName: string): Boolean; virtual; abstract;
  procedure InternalDatabaseCreate(const ADatabaseName: string); virtual; abstract;
  procedure InternalDatabaseDrop(const ADatabaseName: string); virtual; abstract;
  function GetDAOClass(): TibeDALQueryClass; virtual; abstract;
  function GetIsSupportedSequence(): Boolean; virtual; abstract;
  function GetIsSupportedIentityAfterInsert(): Boolean; virtual; abstract;
  function GetIsSupportedTempTable(): Boolean; virtual; abstract;
  property UserName: string read FUserName write FUserName;
  property Password: string read FPassword write FPassword;
  function GetActualKeepConnection(): Boolean; virtual; abstract;
  procedure SetActualKeepConnection(AKeep: Boolean); virtual; abstract;
  function GetInitialDatabase(): string; virtual;
  function GetActiveDatabase(): string; virtual;
  procedure SetActiveDatabase(const ADatabase: string); virtual;
  property TranStartCount: Integer read FTranStartCount;
public
  constructor Create(aOwner: TComponent); virtual;
  destructor Destroy; override;
  procedure Setup();
  procedure Open; virtual; abstract;
  procedure Close; virtual; abstract;
  procedure DatasetInit(a_ds: TDataSet); virtual;
  function DatabaseExists(const ADatabaseName: string = ''): Boolean;
  procedure DatabaseCreate(const ADatabaseName: string = '');
  procedure DatabaseDrop(const ADatabaseName: string = '');
  property DBMSName: string read FDBMSName;
  property KeepConnection: Boolean read FKeepConnection write SetKeepConnection;
  property TraceEnabled: Boolean read FTraceEnabled write FTraceEnabled;
  property ActualKeepConnection: Boolean read GetActualKeepConnection;
  /// <summary> Transaktionen steuern </summary> 
  function TranStart(): Boolean;
  function TranCommit(): Boolean;
  function TranRollback(aOrigException: Exception = nil): Boolean;
  property TranActive: Boolean Read GetTranActive;
  /// <summary> Data Access Object </summary> 
  property DAOClass: TibeDALQueryClass read GetDAOClass;
  /// <summary> Erzeugt eine neue DAO-Query-Instanz </summary> 
  function DAOQueryCreate: TibeDALQuery;
  /// <summary> Konfektioniert die bergebene Query passend zur Connection </summary> 
  procedure DAOQueryInit(a_daoQuery: TibeDALQuery); virtual;
  property IsSupportedSequence: Boolean read GetIsSupportedSequence;
  property IsSupportedIentityAfterInsert: Boolean read GetIsSupportedIentityAfterInsert;
  property IsSupportedTempTable: Boolean read GetIsSupportedTempTable;
  /// <summary>
  ///  Liest nach einem Insert einen vom DBMS generierten PKValue aus.
  ///  Wenn das DBMS den Zhler selbst setzt, kann die Anwendung nach dem Insert
  ///  den vergebenen PKValue auslesen und verarbeiten.
  ///  Wenn das DBMS diese Funktion nicht untersttzt, wird eine EibeDALConnectionNotSupported
  ///  Exceptopn ausgelst.
  /// </summary>
  function RetrieveIdentityValueAfterInsert(const AName: string = ''): Variant; virtual;
  function RetrieveSequenceValue(const AName: string = ''): Variant; virtual;
  function RetrieveTempTableExists(a_sTableName: WideString): Boolean; virtual; 
  procedure RetrieveTableList(a_slTableNames: TStringList); virtual;
  property InitialDatabase: string read GetInitialDatabase;
  property ActiveDatabase: string read GetActiveDatabase write SetActiveDatabase;
  /// <summary> Konvertieren von Werten in Datenbankabhngige String-Reprsentationen </summary> 
  function CVToString(a_vValue: Variant): ibeWideString; virtual;
  /// <summary> 
  /// Konvertieren von Werten in Datenbankabhngige String-Reprsentationen als 
  /// abstrakte Funktionen. Implementierung erfolgt in den spezialisierten Klassen 
  /// </summary> 
  function CVBoole(a_bValue: Boolean): string; virtual; abstract;
  function CVDate(a_dtDate: TDateTime): string; virtual; abstract;
  function CVDateTime(a_dtDate: TDateTime): string; virtual; abstract;
  function CVFloat(a_nValue: ibeFloat): string; virtual; abstract;
  /// <summary> Datanbankparamerter </summary> 
  property Parameters: TStringList read FParameters write FParameters;
end;

TibeDALConnectionClass = class of TibeDALConnection;

TibeDALConnectionFactory = class
private
protected
public
  class function RetriveConnectionClass(const AConnectionTypeID: string): TibeDALConnectionClass;
  class function CreateInstance(AOwner: TComponent; const AConnectionTypeID: string): TibeDALConnection;
  class procedure RegisterConnetion(const AConnectionTypeID: string; AConnectionClass: TibeDALConnectionClass);
end;

EibeDALConnection = class(EibeException);
EibeDALConnectionNotSupported = class(EibeDALConnection);

const
  gc_sDBMS_ID_MSSQL = 'MSSQL';
  gc_sDBMS_ID_Firebird = 'Firebird';

const
  gc_sConnectionType_ADO_MSSQL = 'ADO-MSSQL';
  gc_sConnectionType_FB_IB = 'Firebird-IB';
  
implementation

constructor TibeDALConnection.Create(aOwner: TComponent);
begin
  inherited Create;

  FTranStartCount := 0;
  FParameters := TStringList.Create;

  InternalCreate();
  DoAfterInternalCreate();
end;

destructor TibeDALConnection.Destroy;
begin
  inherited Destroy;

  ibeFreeAndNil(FParameters);
end;

procedure TibeDALConnection.Setup();
begin
  KeepConnection := SameText(Parameters.Values['KeepConntion'], 'True');
  TraceEnabled := SameText(Parameters.Values['SQLTrace'], 'True');
  FDBMSName := Parameters.Values['DBMS'];

  InternalSetup();
end;

procedure TibeDALConnection.InternalCreate(); 
begin
end;

procedure TibeDALConnection.DoAfterInternalCreate(); 
begin
end;

function TibeDALConnection.DAOQueryCreate: TibeDALQuery;
begin
  Result := DAOClass.Create;

  DAOQueryInit(Result);
end;

procedure TibeDALConnection.DAOQueryInit(a_daoQuery: TibeDALQuery);
begin
  InternalDAOQueryInit(a_daoQuery);
end;

function TibeDALConnection.RetrieveIdentityValueAfterInsert(const AName: string = ''): Variant;
begin
  raise EibeDALConnectionNotSupported.Create('Methode RetrieveIdentityValue() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

function TibeDALConnection.RetrieveSequenceValue(const AName: string = ''): Variant;
begin
  raise EibeDALConnectionNotSupported.Create('Methode RetrieveSequenceValue() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

function TibeDALConnection.RetrieveTempTableExists(a_sTableName: WideString): Boolean;
begin
  raise EibeDALConnectionNotSupported.Create('Methode RetrieveTempTableExists() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

procedure TibeDALConnection.RetrieveTableList(a_slTableNames: TStringList);
begin
  raise EibeDALConnectionNotSupported.Create('Methode RetrieveTableList() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

function TibeDALConnection.GetInitialDatabase(): string;
begin
  raise EibeDALConnectionNotSupported.Create('Methode GetInitialDatabase() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

function TibeDALConnection.GetActiveDatabase(): string; 
begin
  raise EibeDALConnectionNotSupported.Create('Methode GetActiveDatabase() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

procedure TibeDALConnection.SetActiveDatabase(const ADatabase: string);
begin
  raise EibeDALConnectionNotSupported.Create('Methode SetActiveDatabase() wird von Connectionklasse ' + ClassName + ' nicht untersttzt');
end;

procedure TibeDALConnection.DatasetInit(a_ds: TDataSet);
begin
  // im Standard gibt es nichts zu tun 
end;

function TibeDALConnection.DatabaseExists(const ADatabaseName: string): Boolean;
begin
  Result := InternalDatabaseExists(ADatabaseName);
end;

procedure TibeDALConnection.DatabaseCreate(const ADatabaseName: string); 
begin
  InternalDatabaseCreate(ADatabaseName);
end;

procedure TibeDALConnection.DatabaseDrop(const ADatabaseName: string); 
begin
  InternalDatabaseDrop(ADatabaseName);
end;

procedure TibeDALConnection.SetKeepConnection(AKeep: Boolean);
begin
  FKeepConnection := AKeep;
  SetActualKeepConnection(KeepConnection);
end;

function TibeDALConnection.GetTranActive(): Boolean; 
begin
  Result := TranStartCount > 0;
end;

function TibeDALConnection.TranStart(): Boolean;
var
  LStarted: Boolean;
begin
  LStarted := False;
  
  if TranStartCount = 0 then
  begin
    SetActualKeepConnection(True); // whrend die Transaktion luft darf die 
                                   // Connection nicht aufgehoben werden, da sonst 
                                   // die Transaktion mit einem impliziten Rollback 
                                   // automatisch abgebrochen wird zumindet bei MSSQL 
    InternalTranStart();
    LStarted := True;
  end;

  Inc(FTranStartCount);

  Result := LStarted;
end;

function TibeDALConnection.TranCommit(): Boolean;
var
  LCommitted: Boolean;
begin
  LCommitted := False;
  
  if TranStartCount > 0 then
  begin
    if TranStartCount = 1 then
    begin
      try
        InternalTranCommit();
        LCommitted := True;
      finally
        SetActualKeepConnection(KeepConnection);  // auf den eingestellten KeepConnection
                                                  // Zustand zurckstellen 
      end;
    end;

    Dec(FTranStartCount);
  end;

  Result := LCommitted;
end;

function TibeDALConnection.TranRollback(AOrigException: Exception = nil): Boolean;
var
  LRolledBack: Boolean;
begin
  LRolledBack := False;
  
  if TranStartCount > 0 then
  begin
    if TranStartCount = 1 then
    begin
      try
        InternalTranRollback();
        LRolledBack := True;
      finally
        SetActualKeepConnection(KeepConnection);  // auf den eingestellten KeepConnection 
                                                  // Zustand zurckstellen 
      end;
    end;

    Dec(FTranStartCount);
  end;

  Result := LRolledBack;
end;

function TibeDALConnection.CVToString(a_vValue: Variant): ibeWideString;
begin
  case VarType(a_vValue) Of
    varEmpty, 
    varNull:      Result := 'null';
    varOleStr, 
    varUString, 
    varString:    Result := AnsiQuotedStr(VarToWideStr(a_vValue), '''');
    varDate:      Result := '''' + CVDateTime(a_vValue) + '''';
    varSingle, 
    varDouble, 
    varCurrency:  Result := CVFloat(a_vValue);
    varBoolean:   Result := CVBoole(a_vValue);
  Else            Result := a_vValue;
  End;
End;

var
  m_slRegisteredConnections: TStringList;

class function TibeDALConnectionFactory.RetriveConnectionClass(const aConnectionTypeID: string): TibeDALConnectionClass;
var
  idx: Integer;
  connectionclass: TibeDALConnectionClass;
begin
  if Assigned(m_slRegisteredConnections) then
  begin
    idx := m_slRegisteredConnections.IndexOf(aConnectionTypeID);

    if idx >= 0 then
    begin
      connectionclass := TibeDALConnectionClass(m_slRegisteredConnections.Objects[idx]);
    end else
    begin
      raise Exception.Create('TibeDALConnectionFactory.RetriveConnectionClass(): die Connection-Klasse "'
                           + aConnectionTypeID + '" ist nicht registriert');
    end;
  end else
  begin
    raise Exception.Create('TibeDALConnectionFactory.RetriveConnectionClass(): es ist noch keine Connection-Klasse registriert');
  end;

  Result := connectionclass;
end;

class function TibeDALConnectionFactory.CreateInstance(aOwner: TComponent; const aConnectionTypeID: string): TibeDALConnection;
begin
  Result := RetriveConnectionClass(aConnectionTypeID).Create(aOwner);
end;

class procedure TibeDALConnectionFactory.RegisterConnetion(const aConnectionTypeID: string; aConnectionClass: TibeDALConnectionClass);
begin
  if not Assigned(m_slRegisteredConnections) then
  begin
    m_slRegisteredConnections := TStringList.Create;
  end;

  m_slRegisteredConnections.AddObject(aConnectionTypeID, TObject(aConnectionClass));
end;

initialization

finalization

FreeAndNil(m_slRegisteredConnections);

end.
