// ============================================================================
//
// Unit:    ibeDataObject
// Inhalt:  Formular
// Aufgabe: DataObject Basisklasse
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeDataObject;

interface

uses
  SysUtils, Variants, Classes, DB,
  ibeString, ibeVariant,
  ibeDALConnection, ibeDALQuery, ibeMemTable,
  ibeDataObjectTypes, ibeDataObjectDefinition, ibeDataObjectFields;

type

TibeDataObject = class;

TibeDataObject = class
private
  FConnection: TibeDALConnection;
  FDODef: TibeDataObjectDefinition;
  FQuery: TibeDALQuery;
  FdatEdit: TibeMemTable;
  FDOClassName: string;
  FState: TibeDataObjectState;
  FFieldList: TibeDataObjectInternalFieldList;
  FPublicField: TibeDataObjectField;
  FTraceEnabled: Boolean;
  procedure Init();
  procedure InitQuery(AQuery: TibeDALQuery);
  procedure InitDatasetFields(ADataSet: TDataSet);
  procedure InitFieldList();
  function GetRecordCount(): Integer;
  function GetFieldCount(): Integer;
  function GetField(const AFieldName: string): TibeDataObjectField;
  function GetFieldByIndex(AIndex: Integer): TibeDataObjectField;
  function GetPKName(): string;
  function GetPKValue(): Variant;
  procedure OnFieldValueChanged(AField: TibeDataObjectInternalField);
  function GetDefinitionFileName(): TFileName;
protected
  procedure InitNewRecord(); virtual;
  procedure AppendToTrace(const AText: String; const AComment: string = '');
  procedure InternalFieldValueChanged(const AFieldName: string; AValue: Variant); virtual;
  function DoBeforePost(): Boolean; virtual;
  function DoAfterPost(): Boolean; virtual;
  procedure DoBeforeDelete(APKValue: Variant); virtual;
  procedure DoAfterDelete(APKValue: Variant); virtual;
  function RetrieveCustomPKValue(): Variant; virtual;
  property DOClassName: string read FDOClassName;
  property DODef: TibeDataObjectDefinition read FDODef;
  property Query: TibeDALQuery read FQuery;
  property FieldList: TibeDataObjectInternalFieldList read FFieldList;
  function GetDefinitionFilePath(): TFileName; virtual;
  property DefinitionFilePath: TFileName read GetDefinitionFilePath;
  property DefinitionFileName: TFileName read GetDefinitionFileName;
public
  constructor Create(AConnection: TibeDALConnection); virtual;
  destructor Destroy(); override;
  property Connection: TibeDALConnection read FConnection;
  procedure Refresh();
  procedure Insert();
  procedure Edit();
  function Post(): Boolean;
  procedure Delete(); overload;
  procedure Delete(APKValue: Variant); overload;
  ///  <summary> DeleteAll() lscht alle im DO geladenen Datenstze </summary>
  ///  <returns> Anzahl gelschter Datenstze </returns>
  function DeleteAll(): Integer;
  function First(): Boolean;
  function Next(): Boolean;
  function Prior(): Boolean;
  procedure Last();
  function BOF(): Boolean;
  function EOF(): Boolean;
  function Fetch(ANoExceptionOnCountMismatch: Boolean = False): Boolean; overload;
  function Fetch(APKValue: Variant; ANoExceptionOnCountMismatch: Boolean = False): Boolean; overload;
  procedure FilterClear();
  function FilterBuild(const AFieldName: string; AFilterValue: Variant): string; overload;
  function RetrieveSQLTextForFieldValue(const AFieldName: string; AValue: Variant; out AValueAsSQLText: string): Boolean;
  function FilterBuild(const AFieldName: string; AFilterValue: Variant; AOpeator: string): string; overload;
  procedure FilterAdd(const AFilter: string); overload;
  procedure FilterAdd(const AFieldName: string; AFilterValue: Variant); overload;
  procedure FilterAdd(const AFieldName: string; AFilterValue: Variant; AOpeator: string); overload;
  property PKName: string read GetPKName;
  property PKValue: Variant read GetPKValue;
  property State: TibeDataObjectState read FState;
  property RecordCount: Integer read GetRecordCount;
  property FieldCount: Integer read GetFieldCount;
  property Field[const AFieldName: string]: TibeDataObjectField read GetField;
  property FieldByIndex[AIndex: Integer]: TibeDataObjectField read GetFieldByIndex;
  procedure DataCopyToDataset(ADat: TibeMemTable);
  property TraceEnabled: Boolean read FTraceEnabled write FTraceEnabled;
end;

TibeDataObjectClass = class of TibeDataObject;

implementation

uses
  ibeAppServiceProviderDB;

constructor TibeDataObject.Create(AConnection: TibeDALConnection); 
begin
  inherited Create();
  FDODef := TibeDataObjectDefinition.Create();
  FFieldList := TibeDataObjectInternalFieldList.Create();
  FPublicField := TibeDataObjectField.Create();
  FQuery := AConnection.DAOQueryCreate();
  FdatEdit := TibeMemTable.Create(nil);

  FState := dosNone;
  FConnection := AConnection;
  TraceEnabled := Connection.TraceEnabled;
  FDOClassName := Copy(ClassName, 2, Length(ClassName)); // vom Klassennamen das fhrende "T" abschneiden

  Init();
end;

destructor TibeDataObject.Destroy(); 
begin
  inherited Destroy();

  FreeAndNil(FDODef);
  FreeAndNil(FFieldList);
  FreeAndNil(FPublicField);
  FreeAndNil(FQuery);
  FreeAndNil(FdatEdit);
end;

procedure TibeDataObject.Init();
begin
  DODef.InitDefinition(DOClassName, DefinitionFileName);

  InitQuery(Query);
end;

procedure TibeDataObject.InitDatasetFields(ADataSet: TDataSet);
var
  i: Integer;
  LViewField: TibeDODefViewField;
  LDBField: TField;
begin
  if ADataSet.Tag <> DODef.ActiveView.FieldModifyTag then
  begin
    // Mit dem genderten ModifyTag signalisiert die View dass sie gendert wurde.
    // Die Feldlisten von DataObject und Abfrage mssen in diesem Fall neu
    // aufgebaut werden.
    ADataSet.Fields.Clear();
    ADataSet.FieldDefs.Clear();
    //Query.DataSet.FieldDefs.Update();

    FieldList.Clear();
    for i := 0 to DODef.ActiveView.FieldCount - 1 do
    begin
      LViewField := DODef.ActiveView.Field[i];
      LDBField := LViewField.AddFieldToDataset(ADataSet);
      FieldList.AddField(LViewField, LDBField, OnFieldValueChanged);
    end;

    // Das ModifyTag der View fr nachfolgende Vergleiche merken
    ADataSet.Tag := DODef.ActiveView.FieldModifyTag;
  end;
end;

procedure TibeDataObject.InitFieldList();
var
  i: Integer;
  LViewField: TibeDODefViewField;
begin
  if FieldList.DefFieldsModifiedTag <> DODef.ActiveView.FieldModifyTag then
  begin
    // Mit dem genderten ModifyTag signalisiert die View dass sie gendert wurde.
    // Die Feldliste vom DataObject muss in diesem Fall neu aufgebaut werden.
    FieldList.Clear();

    for i := 0 to DODef.ActiveView.FieldCount - 1 do
    begin
      LViewField := DODef.ActiveView.Field[i];
      FieldList.AddField(LViewField, nil, OnFieldValueChanged);
    end;

    // Das ModifyTag der View fr nachfolgende Vergleiche merken
    FieldList.DefFieldsModifiedTag := DODef.ActiveView.FieldModifyTag;
  end;
end;

procedure TibeDataObject.InitQuery(AQuery: TibeDALQuery);
begin
  AQuery.Close();
  AQuery.SQL := DODef.SQLSelect;
  AppendToTrace(AQuery.SQL);

  //InitDatasetFields(AQuery.DataSet);
  InitFieldList();
end;

procedure TibeDataObject.Refresh();
begin
  InitQuery(Query);

  Query.Open();
  FieldList.BindDataSetFields(Query.DataSet);

  //FFieldList.ExchangeDataSet(Query.DataSet);

  FState := dosBrowse;
  
  First();
end;

procedure TibeDataObject.Insert();
begin
  if not Query.DataSet.Active then
  begin
    Refresh();
  end;

  FdatEdit.Close();
//  InitDatasetFields(FdatEdit);
//  FFieldList.ExchangeDataSet(FdatEdit);
  InitFieldList();
  FieldList.AddFieldsToDataset(FdatEdit);
  FieldList.BindDataSetFields(FdatEdit);
  FdatEdit.Open();
  FdatEdit.Append();

  FState := dosInsert;

  InitNewRecord();
end;

procedure TibeDataObject.InitNewRecord(); 
begin
end;

procedure TibeDataObject.Edit();
begin
  if Query.DataSet.Locate(PKName, PKValue, []) then
  begin
    FdatEdit.Close();
//    InitDatasetFields(FdatEdit);
//    FFieldList.ExchangeDataSet(FdatEdit);
    InitFieldList();
    FieldList.AddFieldsToDataset(FdatEdit);
    FieldList.BindDataSetFields(FdatEdit);
    FdatEdit.LoadFromDataset(Query.DataSet, False);
    FdatEdit.Edit();

    FState := dosEdit;
  end else
  begin
    raise EibeDataObjectDefinition.Create('DataObject "' + DOClassName + '": Edit nict mglich, der Datensatz ist nicht geladen');
  end;
end;

function TibeDataObject.Post(): Boolean;
var
  i: Integer;
  LPKValue: Variant;
  LRecordPosted: Boolean;
  LValueAsSQLText: string;
  LCtrlTran: Boolean;
begin
  LRecordPosted := False;
  LCtrlTran := Connection.TranStart();
  try
    DODef.ClearPostParams();

    LRecordPosted := DoBeforePost();
  
    for i := 0 to FFieldList.Count - 1 do
    begin
      if FFieldList[i].Modified                                           // nur genderte Felder aktualisieren
        and not SameText(FFieldList[i].FieldName, DODef.MainTablePKName)  // hier nicht das PK-Feld verarbeiten
        and SameText(FFieldList[i].TblAlias, DODef.MainTableAlias)        // nur wenn das Feld aus der MainTable kommt
      then begin
        if not RetrieveSQLTextForFieldValue(FFieldList[i].FieldName, FFieldList[i].Value, LValueAsSQLText) then
        begin
          raise EibeDataObject.Create('DataObject "' + DOClassName + '": Parameter fr Feld "' + FFieldList[i].FieldName + '" kann nicht angelegt werden, das Feld wurde nicht gefunden');
        end else
        begin
          DODef.AddPostParam(FFieldList[i].FieldName, LValueAsSQLText);
        end;
      end;  
    end;

    if DODef.PostParamCount > 0 then
    begin
      if State = dosInsert then
      begin
        if DODef.MainTableKKValueMode = pkmCustom then
        begin
          LPKValue := RetrieveCustomPKValue();
        
          if not VarHasValue(LPKValue) then
          begin
            raise EibeDataObject.Create('Datenobjekt ' + DOClassName + '.Insert(): fr das PK-Feld ' + PKName + ' ist kein Wert angegeben');
          end;
      
          if not RetrieveSQLTextForFieldValue(PKName, LPKValue, LValueAsSQLText) then
          begin
            raise EibeDataObject.Create('DataObject "' + DOClassName + '": Parameter fr Feld "' + PKName + '" kann nicht angelegt werden, das Feld wurde nicht gefunden');
          end else
          begin
            DODef.AddPostParam(PKName, LValueAsSQLText);
          end;
        end;
      
        Query.SQL := DODef.SQLInsert;
      end else if State = dosEdit then
      begin
        FilterClear();
        DODef.PKFilterAdd(PKValue, True);
        Query.SQL := DODef.SQLUpdate;
      end;
  
      AppendToTrace(Query.SQL);
  
      try
        LPKValue := PKValue;
        Query.Execute();
        LRecordPosted := True;
        if State = dosInsert then
        begin
          if DODef.MainTableKKValueMode = pkmSerial then
          begin
            LPKValue := Connection.RetrieveIdentityValueAfterInsert(DODef.MainTablePKSequenceName);
          end;
        end;
        Fetch(LPKValue);
      finally
        FState := dosNone;
      end;
    end;

    if LRecordPosted then
    begin
      LRecordPosted := DoAfterPost();
    end;
    
    Connection.TranCommit();
  except
    on E: Exception do
    begin
      Connection.TranRollback(E);
      raise;
    end;
  end;

  Result := LRecordPosted;
end;

procedure TibeDataObject.Delete(APKValue: Variant); 
var
  LCtrlTran: Boolean;
begin
  LCtrlTran := Connection.TranStart();
  try
    Fetch(APKValue);
    
    DoBeforeDelete(APKValue);
    
    FilterClear();
    DODef.PKFilterAdd(APKValue, True);
    Query.SQL := DODef.SQLDelete;
  
    AppendToTrace(Query.SQL);
  
    Query.Execute();

    DoAfterDelete(APKValue);
    
    Connection.TranCommit();
  except
    on E: Exception do
    begin
      Connection.TranRollback(E);
      raise;
    end;
  end;
end;

function TibeDataObject.DeleteAll(): Integer;
var
  i: Integer;
  LCtrlTran: Boolean;
  LPKArray: Variant;
//  LDat: TibeMemTable;
begin
  Result := 0;
  // Variantes Array zur Aufnahme der selektierten PKValues erstellen
  LPKArray := VarArrayCreate([0, Pred(RecordCount)], varVariant);

  First();
  i := 0;
  // PKValues aller geladenen Datenstze in das Array eintragen
  while not EOF do
  begin
    LPKArray[i] := PKValue;
    Inc(i);
    Next();
  end;

  LCtrlTran := Connection.TranStart();
  try
    // ber alle Arrayelemente iterieren und mit den hinterlegten PKValues die
    // Datenstze lschen
    for i := 0 to VarArrayHighBound(LPKArray, 1) do
    begin
      Delete(LPKArray[i]);
      Inc(Result);
    end;
    
    Connection.TranCommit();
  except
    on E: Exception do
    begin
      Connection.TranRollback(E);
      raise;
    end;
  end;

//  Result := 0;
//  LDat := TibeMemTable.Create(nil);
//  try
//    LDat.LoadFromDataset(Query.DataSet, True);
//
//    LCtrlTran := Connection.TranStart();
//    try
//      // ber alle Arrayelemente iterieren und mit den hinterlegten PKValues die
//      // Datenstze lschen
//      LDat.First();
//      while not LDat.Eof do
//      begin
//        Delete(LDat.FieldByName(PKName).Value);
//        Inc(Result);
//        LDat.Next();
//      end;
//    
//      Connection.TranCommit();
//    except
//      on E: Exception do
//      begin
//        Connection.TranRollback(E);
//        raise;
//      end;
//    end;
//  finally
//    FreeAndNil(LDat);
//  end;
end;

procedure TibeDataObject.Delete();
begin
  Delete(PKValue);
end;

function TibeDataObject.First(): Boolean;
begin
  Query.DataSet.First();

  Result := not Query.DataSet.Eof;

  if Result then
  begin
    FieldList.InitFieldsFromDataset();
  end;
end;

function TibeDataObject.Next(): Boolean;
begin
  Query.DataSet.Next();

  Result := not Query.DataSet.Eof;

  if Result then
  begin
    FieldList.InitFieldsFromDataset();
  end;
end;

function TibeDataObject.Prior(): Boolean;
begin
  Query.DataSet.Prior();

  Result := not Query.DataSet.Bof;

  if Result then
  begin
    FieldList.InitFieldsFromDataset();
  end;
end;

procedure TibeDataObject.Last();
begin
  Query.DataSet.Last();

  FieldList.InitFieldsFromDataset();
end;

function TibeDataObject.BOF(): Boolean;
begin
  Result := Query.DataSet.Bof;
end;

function TibeDataObject.EOF(): Boolean;
begin
  Result := Query.DataSet.Eof;
end;

function TibeDataObject.Fetch(ANoExceptionOnCountMismatch: Boolean): Boolean; 
begin
  Result := Fetch(PKValue);
end;

function TibeDataObject.Fetch(APKValue: Variant; ANoExceptionOnCountMismatch: Boolean): Boolean; 
begin
  FilterClear();
  FilterAdd(PKName, APKValue);
  Refresh();

  Result := RecordCount = 1;
  if not (Result or ANoExceptionOnCountMismatch) then
  begin
    raise EibeDOFetchCountMismatch.Create('Fehler in Datenobjekt ' + DOClassName + ': Fetch() muss genau einen Datensatz liefern, geladen wurden ' + IntToStr(RecordCount));
  end;
end;

procedure TibeDataObject.FilterClear();
begin
  DODef.FilterClear();
end;

function TibeDataObject.FilterBuild(const AFieldName: string; AFilterValue: Variant): string;
begin
  Result := FilterBuild(AFieldName, AFilterValue, '=');
end;

function TibeDataObject.RetrieveSQLTextForFieldValue(const AFieldName: string; AValue: Variant; out AValueAsSQLText: string): Boolean;
var
  LNativeValue: Variant;
  LInternalField: TibeDataObjectInternalField;
  LTblField: TibeDODefTableField;
begin
  Result := False;
  LInternalField := FFieldList.FieldByName[AFieldName];

  if Assigned(LInternalField) then
  begin
    // das Feld ist in der Feldliste vom DO bekannt. Damit den nativen Feldwert bilden
    LNativeValue := LInternalField.ConvertToTableFieldValue(AValue);
    Result := True;
  end else
  begin
    // das Feld ist in der DO-Feldliste nicht enthalten. Auf die Tabellenfelder 
    // in der DO-Definition ausweichen
    LTblField := DODef.FieldByName[AFieldName];
    if Assigned(LTblField) then
    begin
      LNativeValue := LTblField.ConvertToTableFieldValue(AValue);
      Result := True;
    end;
  end;
  
  AValueAsSQLText := Connection.CVToString(LNativeValue);
end;

function TibeDataObject.FilterBuild(const AFieldName: string; AFilterValue: Variant; AOpeator: string): string; 
var
  LValueAsSQLText: string;
  LValueString: Variant;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObject.Create('Datenobjekt ' + DOClassName + ': FilterBuild() kann mit leerem Feldnamen nicht aufgerufen werden');
  end else
  begin
    if VarHasValue(AFilterValue) then
    begin
      if not RetrieveSQLTextForFieldValue(AFieldName, AFilterValue, LValueAsSQLText) then
      begin
        raise EibeDataObject.Create('Datenobjekt ' + DOClassName + ': FilterBuild() fr Feld "' + AFieldName + '" nicht mglich, das Feld ist in der View + "' + DODef.ActiveView.Name + '" nicht bekannt');
      end else
      begin
        LValueString := LValueAsSQLText;
      end;
    end else
    begin
      LValueString := Null;
    end;
  end;
  
  Result := DODef.FilterBuild(AFieldName, LValueString, AOpeator);
end;

procedure TibeDataObject.FilterAdd(const AFilter: string); 
begin
  DODef.FilterAdd(AFilter);
end;

procedure TibeDataObject.FilterAdd(const AFieldName: string; AFilterValue: Variant); 
begin
  FilterAdd(AFieldName, AFilterValue, '=');
end;

procedure TibeDataObject.FilterAdd(const AFieldName: string; AFilterValue: Variant; AOpeator: string); 
begin
  FilterAdd(FilterBuild(AFieldName, AFilterValue, AOpeator));
end;

procedure TibeDataObject.OnFieldValueChanged(AField: TibeDataObjectInternalField);
begin
  InternalFieldValueChanged(AField.FieldName, AField.Value);
end;

procedure TibeDataObject.InternalFieldValueChanged(const AFieldName: string; AValue: Variant); 
begin
end;

function TibeDataObject.DoBeforePost(): Boolean;
begin
  Result := False;
end;

function TibeDataObject.DoAfterPost(): Boolean;
begin
  Result := True;
end;

procedure TibeDataObject.DoBeforeDelete(APKValue: Variant); 
begin
end;

procedure TibeDataObject.DoAfterDelete(APKValue: Variant); 
begin
end;

function TibeDataObject.RetrieveCustomPKValue(): Variant; 
begin
  Result := Field[PKName].Value;
end;

function TibeDataObject.GetDefinitionFilePath(): TFileName;
begin
  Result := AddBackslash(RetrieveASP.CfgPath) + 'DOD\';
end;

function TibeDataObject.GetDefinitionFileName(): TFileName;
var
  LDefinitionFileFound: Boolean;
  LDefnitionFileName: TFileName;
  LDOClassName: string;
  LDOFileClass: TibeDataObjectClass;
begin
  LDefinitionFileFound := False;
  LDefnitionFileName := '';
  LDOFileClass := TibeDataObjectClass(ClassType);
  LDOClassName := Copy(ClassName, 2, Length(ClassName)); // vom Klassennamen das fhrende "T" abschneiden

  while not LDefinitionFileFound and (LDOFileClass <> TibeDataObject) do
  begin
    LDefnitionFileName := AddBackslash(DefinitionFilePath) + LDOClassName + '.dod';
    LDefinitionFileFound := FileExists(LDefnitionFileName);
    LDOFileClass := TibeDataObjectClass(LDOFileClass.ClassParent());
    LDOClassName := Copy(LDOFileClass.ClassName, 2, Length(LDOFileClass.ClassName)); // vom Klassennamen das fhrende "T" abschneiden
  end;

  if not LDefinitionFileFound then
  begin
    raise EibeDataObjectDefinition.Create(Self, 'Definition fr Datenobjekt ' + ClassName + ' kann nicht initialisiert werden, es wurde keine Definitionsdatei gefunden.');
  end;

  Result := LDefnitionFileName;
end;

function TibeDataObject.GetRecordCount(): Integer;
//var
//  LBookmark: TBookmark;
begin
  Result := Query.RecordCount;
//  LBookmark := Query.DataSet.GetBookmark();
//  try
//    Query.DataSet.Last();
//    Result := Query.DataSet.RecordCount;
//    Query.DataSet.GotoBookmark(LBookmark);
//  finally
//    Query.DataSet.FreeBookmark(LBookmark);
//  end;
end;

function TibeDataObject.GetFieldCount(): Integer;
begin
  Result := FFieldList.Count;
end;

function TibeDataObject.GetField(const AFieldName: string): TibeDataObjectField;
begin
  if AFieldName = '' then
  begin
    raise EibeDataObject.Create('DataObject "' + DOClassName + '": auf Feld mit leerem Feldnamen kann nicht zugegriffen werden');
  end else
  begin
    FPublicField.SetInternalField(FieldList.FieldByName[AFieldName]);

    if not FPublicField.IsInternalFieldAssigned then
    begin
      raise EibeDataObject.Create('DataObject "' + DOClassName + '": Feld mit Feldnamen "' + AFieldName + '" ist nicht vorhanden');
    end;

    Result := FPublicField;
  end;
end;

function TibeDataObject.GetFieldByIndex(AIndex: Integer): TibeDataObjectField;
begin
  FPublicField.SetInternalField(FieldList[AIndex]);

  Result := FPublicField;
end;

function TibeDataObject.GetPKName(): string;
begin
  Result := DODef.MainTablePKName;
end;

function TibeDataObject.GetPKValue(): Variant;
var
  LField: TibeDataObjectInternalField;
begin
  LField := FFieldList.FieldByName[DODef.MainTablePKName];
  if not Assigned(LField) then
  begin
    raise EibeDataObject.Create('DataObject "' + DOClassName + '": PK-Feld mit Feldnamen "' + DODef.MainTablePKName + '" ist nicht vorhanden');
  end else
  begin
    //Result := LField.ConvertToTableFieldValue(LField.Value);
    Result := LField.Value;
  end;
end;

procedure TibeDataObject.DataCopyToDataset(ADat: TibeMemTable);
var
  LField: TibeDataObjectInternalField;
begin
  ADat.Close();
  DODef.ActiveView.CreateDatasetStructure(ADat);
  ADat.Open();

  First();
  while not EOF() do
  begin
    ADat.Append();
    for LField in FieldList do
    begin
      ADat.Fields[LField.DBFieldIndex].Value := LField.Value;
    end;
    ADat.Post();      
    Next();
  end;
end;

procedure TibeDataObject.AppendToTrace(const AText: String; const AComment: string = '');
var
  LText: string;
  LFileMode: Word;
  LStream: TFileStream;
begin
  try
    if (RetrieveASP.DOTraceEnabled or TraceEnabled) and (Length(AText) > 0) then
    begin
      if not FileExists(RetrieveASP.DOTraceFileName) then
      begin
        LFileMode := fmCreate or fmShareDenyWrite;
      end else
      begin
        LFileMode := fmOpenWrite or fmShareDenyWrite;
      end;
      
      LStream := TFileStream.Create(RetrieveASP.DOTraceFileName, LFileMode);
      try
        LStream.Position := LStream.Size;
        LText := #13#10                                                      // Zeilenumbruch
               + '-- '                                                       // SQL Kommentar einleiten
               + FormatDateTime('yyyy-mm-dd hh:nn:ss.zzz', Now()) + ' '      // Timestamp
               + DOClassName                                                 // Klassenname vom DO
               + '(0x' + IntToHex(Integer(Self), 2 * SizeOf(Integer)) + ')'  // Adresse der DO-Instanz
               + #13#10;                                                     // Zeilenumbruch
        LStream.Write(LText[1], Length(LText) * SizeOf(LText[1]));

        if AComment <> '' then
        begin
          LText := '-- ' + AComment;
          LStream.Write(LText[1], Length(LText) * SizeOf(LText[1]));
        end;
        
        LStream.Write(AText[1], Length(AText) * SizeOf(AText[1]));
      finally
        FreeAndNil(LStream);
      end;
    end;
  except
    // nix machen
  end;
end;

end.
