// ============================================================================
//
// Unit:    ibeFileUtil
// Inhalt:  Formular
// Aufgabe: Funktionsbibliothek mit Funktionen zum Umgang Dateien und 
//          Verzeichnissen
//
//   (c) by Klaus Etscheid since 2016
// ============================================================================
unit ibeFileUtil;

interface

Uses
	SysUtils, Classes, Windows,
	ibeOSProcedures, ibeString, ibeOSConst;

procedure DeleteFile(const a_sFileName: TFileName);
Function GetValidFilename( Const s: String ): String;
procedure RetrieveFileList(const a_sPath: TFileName; a_slFiles: TStringList);

{ ------------------------------------------------------------------------------------ }
{ GetGenerationFileName():                                                             }
{ ------------------------                                                             }
{ FileName wird in die Bestandteile Pfad, Name und Erweiterung zerlegt. Danach wird    }
{ im Pfad nach Generations-Dateien gesucht. Der Dateiname entspricht folgender         }
{ Syntax:                                                                              }
{ Dateiname in FileName: [<path>\]<dateiname>[.<ext>]                                  }
{ Generationsdateiname:  [<path>\]<dateiname>-n[.<ext>] mit n als Generationsnummer.   }
{ Sind im Verzeichnis bereits MaxGenerations Generationen vorhanden, wir die erste     }
{ Generation (n = 1) gelscht, anschlieend die Generationsnummern aller anderen       }
{ Generationen um 1 verringert und danach der der Generations-Name mit  MaxGeneration  }
{ gebildet.                                                                            }
{ ------------------------------------------------------------------------------------ }
Function GetGenerationFileName( Const FileName: String; MaxGenerations: Integer;
																GenerationDelimiter: Char = '-' ): String;

Function ExtractFileJustName( Const FileName: String ): String;

Procedure CreatePath( Const Path: String );
Function IsNetworkPath( Const Path: String ): Boolean;

Function GetTempPath: String;

Function GetEInOutErrorText( ErrorCode: Integer ): String;

Type

TibeFileAttribute = (ifaReadOnly, ifaHidden, ifaSystem, ifaDirectory, ifaArchive,
										 ifaNormal, ifaTemporary, ifaCompressed, ifaOffline, ifaNone );
TibeFileAttributes = Set Of TibeFileAttribute;

Function WinFileAttribureToFileAttributes( attrib: LongWord ): TibeFileAttributes;

Type

TibeFileCreateMode = (ifcmOverwrite, ifcmAppend, ifcmGenerations, ifcmNone);

implementation

uses
  ibeException;

procedure RetrieveFileList(const a_sPath: TFileName; a_slFiles: TStringList);
var iFindResult: Integer;
  sr: TSearchRec;
begin
	iFindResult := SysUtils.FindFirst(a_sPath, faAnyFile - faDirectory, sr);

  try
    while iFindResult = 0 do begin
      a_slFiles.Add(sr.Name);

      iFindResult := SysUtils.FindNext(sr);
    end;
  finally
    SysUtils.FindClose(sr);
  end;
end;

procedure DeleteFile(const a_sFileName: TFileName);
begin
  if not SysUtils.DeleteFile(a_sFileName) then
  begin
    raise EibeException.Create('Datei "' + a_sFileName + '" konnte nicht gelscht werden.');
  end;
end;

Function GetValidFilename( Const s: String ): String;
Begin
	Result := FilterString( s, '\/.<>|:' );
End;

Function GetGenerationFileName( Const FileName: String; MaxGenerations: Integer;
	GenerationDelimiter: Char ): String;
Var s, s1, Masq, Path, Name, Ext: String;
	i, FindResult, Generation: Integer;
	sr: TSearchRec;
	sl: TStringList;
Begin
	sl := TStringList.Create;

	Path := ExtractFilePath( FileName );
	Ext := ExtractFileExt( FileName );
	Name := ExtractFileJustName( FileName );

	Masq := AddBackslash(Path ) + Name + GenerationDelimiter + '*' + Ext;
	FindResult := FindFirst( Masq, faAnyFile, sr );

	While FindResult = 0 Do Begin
		Try
			i := WordCount( sr.Name, [GenerationDelimiter] );
			If i > 1 Then Begin
			{ GenerationDelimiter mindestens einmal gefunden. Jetzt den Teil hinter }
			{ dem letzten GenerationDelimiter extrahieren.                          }
				s := ExtractWord( i, sr.Name, [GenerationDelimiter] );
				s1 := Copy( sr.Name, 1, Length(sr.Name) - Length(s) - 1 );
			{ Die Datei nur aufnehmen, wenn der Teil vor dem letzten GenerationDelimiter }
			{ mit dem vorgegebenen Dateinamen ohne Erweiterung bereinstimmt.            }
				If SameText( s1, Name ) Then Begin
				{ Eine evtl. vorhandene Dateinamenserweiterung abschneiden. }
					WordCount( sr.Name, ['.'] );
					s := ExtractWord( 1, s, ['.'] );
				{ s darf jetzt nur noch die Gernerationszahl enthalten. Die Generationszahl }
				{ ermitteln.                                                                }
					Generation := StrToInt( s );
				{ Jetzt ist alles bekannt. Den Dateinamen mitsamt dem vorgestellten Generation }
				{ in eine Stringliste einfgen. Die Gernration im Object-Pointer merken.      }
					sl.AddObject( Format( '%10.10d %s', [Generation, sr.Name] ), Pointer( Generation ) );
				End;
			End;
		Except
		{ Beim Parsen des Dateinamens sind Fehler aufgetreten, der Dateiname entspricht  }
		{ also nicht der Syntax fr Generationsdateien (der Wildcard '*' darf nur Zahlen }
		{ erfassen). Diese Fehler ignorieren.                                            }
		End;
		FindResult := FindNext( sr );
	End;
	sl.Sort;

{ Die Stringliste enthlt jetzt alle Dateinamen die der Generationensyntax gengen. }
{ Dei Dateinamen sind aufsteigend nacht ihrem Timestamp sortiert. Jetzt die         }
{ ltesten Dateien so lange lschen, bis incl. der neuen Generation maximal         }
{ MaxGenerations Generationen im Verzeichnis liegen werden.                         }
	For i := sl.Count - MaxGenerations DownTo 0 Do Begin
		s := AddBackslash(Path ) + Copy( sl[i], 12, Length( sl[i] ) );
		If Not SysUtils.DeleteFile( PChar(s) )
		Then Begin
			Raise Exception.Create( 'GetGenerationFileName( ' + FileName + ' ): Datei ' + s + ' kann nicht gelscht werden.' );
		End;
		sl.Delete( i );
	End;

{ Jetzt die verbliebenen Dateien nach Ihrer Reihenfolge in der Stringliste mit neuen }
{ Generationsnummern versehen.                                                       }
	Result := AddBackslash(Path ) + Name + GenerationDelimiter + IntToStr(MaxGenerations) + Ext;
	For i := 0 To sl.Count - 1 Do Begin
		s := AddBackslash(Path ) + Copy( sl[i], 12, Length( sl[i] ) );
		s1 := Format( '%s%s%s%d%s', [AddBackslash( Path ), Name, GenerationDelimiter, i + 1, Ext] );
		If Not RenameFile( PChar(s), s1 )
		Then Begin
			Raise Exception.Create( 'GetGenerationFileName( ' + FileName + ' ): Datei ' + s + ' kann nicht nach ' + s1 + ' umbenannt werden.' );
		End;
	End;

{ sl.Count gibt jetzt die hchste Gernerationsnummer an. Mit sl.Count + 1 den Dateinamen bilden. }
	Result := Format( '%s%s%s%d%s', [AddBackslash( Path ), Name, GenerationDelimiter, sl.Count + 1, Ext] );

	sl.Destroy;
End;

Function ExtractFileJustName( Const FileName: String ): String;
Var Ext: String;
Begin
	Result	:= ExtractFileName( FileName );
	Ext			:= ExtractFileExt( Result );
	Result	:= Copy( Result, 1, Length(Result) - Length( Ext ) );
End;

Procedure CreatePath( Const Path: String );
Var i, StartWord, anz: Integer;
	tmppath, srcpath: String;
Begin
	If IsNetworkPath( Path ) Then Begin
		srcpath		:= Copy( Path, 3, Length( Path ) );
		tmppath		:= '\\' + ExtractWord( 1, srcpath, [PathDelim] );
		StartWord	:= 2
	End Else Begin
		tmppath		:= '';
		srcpath		:= Path;
		StartWord	:= 1;
	End;
	anz := WordCount( srcpath, [PathDelim] );

	tmppath := AddBackslash( tmppath ) + ExtractWord( StartWord, srcpath, [PathDelim] );
	For i := StartWord + 1 To anz + 1 Do Begin
		If Not DirectoryExists( tmppath ) Then Begin
			If Not CreateDir( tmppath ) Then Raise Exception.Create( 'CreatePath: Kann Verzeichnis ''' + tmppath + ''' nicht anlegen' );
		End;
		tmppath := tmppath + PathDelim + ExtractWord( i, srcpath, [PathDelim] );
	End;
End;

Function IsNetworkPath( Const Path: String ): Boolean;
Begin
	Result := (Length( Path ) > 2)                	// mindestens 3 Zeichen
				And (Path[1] = '\') And (Path[2] = '\');	// Beginnend mit '\\'
End;

Function GetTempPath: String;
Begin
	Result := ibeOSProcedures.GetTempPath;
End;

Function GetEInOutErrorText( ErrorCode: Integer ): String;
Begin
	Case ErrorCode Of
		0..99:	Result := SysErrorMessage( ErrorCode );
			100:	Result := 'EOF';
			101:	Result := 'Festplatte voll';
			102:	Result := 'Dateivariable nicht zugewiesen';
			103:	Result := 'Datei nicht geffnet';
			104:	Result := 'Datei nicht fr Eingabe geffnet';
			105:	Result := 'Datei nicht fr Ausgabe geffnet';
			106:	Result := 'Ungltige Eingabe';
			107:	Result := 'Datei ist bereits geffnet';
	Else			Result := 'Kein Text fr Fehlernummer ' + IntToStr( ErrorCode );
	End;
End;

{ Konstanten aus windows.pas bernommen. }
{ letzte nderung: 2002-03-20            }
Const
	FILE_ATTRIBUTE_READONLY             = $00000001;
	FILE_ATTRIBUTE_HIDDEN               = $00000002;
	FILE_ATTRIBUTE_SYSTEM               = $00000004;
	FILE_ATTRIBUTE_DIRECTORY            = $00000010;
	FILE_ATTRIBUTE_ARCHIVE              = $00000020;
	FILE_ATTRIBUTE_NORMAL               = $00000080;
	FILE_ATTRIBUTE_TEMPORARY            = $00000100;
	FILE_ATTRIBUTE_COMPRESSED           = $00000800;
	FILE_ATTRIBUTE_OFFLINE              = $00001000;

Function WinFileAttribureToFileAttributes( attrib: LongWord ): TibeFileAttributes;
Begin
	Result := [];

	If (attrib And FILE_ATTRIBUTE_READONLY)		<> 0 Then Result := Result + [ifaReadOnly];
	If (attrib And FILE_ATTRIBUTE_HIDDEN)			<> 0 Then Result := Result + [ifaHidden];
	If (attrib And FILE_ATTRIBUTE_SYSTEM)			<> 0 Then Result := Result + [ifaSystem];
	If (attrib And FILE_ATTRIBUTE_DIRECTORY)	<> 0 Then Result := Result + [ifaDirectory];
	If (attrib And FILE_ATTRIBUTE_ARCHIVE)		<> 0 Then Result := Result + [ifaArchive];
	If (attrib And FILE_ATTRIBUTE_NORMAL)			<> 0 Then Result := Result + [ifaNormal];
	If (attrib And FILE_ATTRIBUTE_TEMPORARY)	<> 0 Then Result := Result + [ifaTemporary];
	If (attrib And FILE_ATTRIBUTE_COMPRESSED)	<> 0 Then Result := Result + [ifaCompressed];
	If (attrib And FILE_ATTRIBUTE_OFFLINE) 		<> 0 Then Result := Result + [ifaOffline];
End;

end.
