// ============================================================================
//
// Unit:    ibeSQLScript
// Inhalt:  Formular
// Aufgabe: Klassenbibliothek zum Ausfhren eines SQL-Scripts
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit ibeSQLScript;
// =============================================================================
// Die Klasse TibeSQLScript instanziert Parser und Worker, bergibt den Worker
// an den Parser und ruft desesn Execute() auf.
// Der Script-Parser liest eine Datei ein, zerlegt diese in einzelne Statements
// und ruft fr jedes Statement den bergebenen Worker auf.
//
// Script Syntax:
// --------------
// Ein Statement beginnt mit der ersten Zeile mit einem nicht White-Space
// Zeichen.
// Ein Statement endet mit der Zeile die mit einem Semikolon endet oder mit 
// einer Zeile, die auer White-Spaces nur den Text "go" enthlt (eine 
// MSSQL-Reminiszenz).
// Kommentare nach den Ende-Tokens sind nicht erlaubt, Kommentare innerhalb vom
// Statement werden dem Text des Statements zugeordnet.
// Nach dem Ende eines Statements wird der Beginn vom nchsten Statement gesucht
// und dieses dann verarbeitet. Es wird bis zum Dateiende oder dem ersten 
// fehlerhaft ausgefhrten Statement fortgefahren.
// =============================================================================
interface

uses
  SysUtils, StrUtils, Classes,
  ibeString, ibeFileUtil, ibeOSProcedures,
  ibeDALConnection, ibeDALQuery;

type

TibeSQLStatement = record
  Index: Integer;
  RowNum: Integer;
  SQLStatement: string;
  ScriptFileName: TFileName;
end;
PibeSQLStatement = ^TibeSQLStatement;

TibeSQLStatementWorker = class
private
  FQuery: TibeDALQuery;
protected
public
  constructor Create(AConnection: TibeDALConnection); virtual;
  destructor Destroy(); override;
  procedure Execute(const AStatement: PibeSQLStatement); virtual; abstract;
end;

TibeSQLStatementWorkerClass = class of TibeSQLStatementWorker;

TibeSQLStatementSaveToFileWorker = class(TibeSQLStatementWorker)
private
  FStream: TFileStream;
  FWriter: TStreamWriter;
protected
public
  destructor Destroy(); override;
  procedure Execute(const AStatement: PibeSQLStatement); override;
end;

/// <summary>
///  TibeSQLScript ldt ein SQL Script, zerlegt es in einzelne Statements und
///  fhrt dieses aus. Der Klasse kann eine Worker-Klasse zugewiesen werden, mit
///  der der Aufrufer die Ausfhrung der Statements steuern kann
/// </summary>
TibeSQLScript = class
strict private
  type
    ///<summary> Schmalspur Worker, macht das, wie er heit: fhrt das bergeene Statement aus </summary>
    TExecWorker = class(TibeSQLStatementWorker)
    private
    protected
    public
      procedure Execute(const AStatement: PibeSQLStatement); override;
    end;
private
  FConnection: TibeDALConnection;
  FStatementWorkerClass: TibeSQLStatementWorkerClass;
  FFileName: TFileName;
  FActivateTransactions: Boolean;
protected
  property Connection: TibeDALConnection read FConnection;
public                                  
  constructor Create(AConnection: TibeDALConnection); virtual;
  procedure Execute();
  property FileName: TFileName read FFileName write FFileName;
  property StatementWorkerClass: TibeSQLStatementWorkerClass read FStatementWorkerClass write FStatementWorkerClass;
  property ActivateTransactions: Boolean read FActivateTransactions write FActivateTransactions;
end;

TibeSQLScriptParser = class
private
protected
public
  procedure Execute(const AFileName: TFileName; AWorker: TibeSQLStatementWorker);
end;

const
  GC_INCLUDE_PREFIX = '--$Include=';

implementation

// =========================== [ TibeSQLScript ] ===============================

constructor TibeSQLScript.Create(AConnection: TibeDALConnection); 
begin
  inherited Create();

  FConnection := AConnection;
  StatementWorkerClass := TExecWorker;  // Default: Statement ausfhren
  ActivateTransactions := True;
end;

procedure TibeSQLScript.Execute();
var
  LCtrllTran: Boolean;
  LScriptParser: TibeSQLScriptParser;
  LStatementWorker: TibeSQLStatementWorker;
begin
  LCtrllTran := ActivateTransactions;
  
  LStatementWorker := StatementWorkerClass.Create(FConnection);
  try
    LScriptParser := TibeSQLScriptParser.Create();
    try
      if ActivateTransactions then
      begin
        LCtrllTran := FConnection.TranStart;
      end;
      
      try
        LScriptParser.Execute(FileName, LStatementWorker);

        if ActivateTransactions then
        begin
          FConnection.TranCommit();
        end;
      except
        on E: Exception do
        begin
          if ActivateTransactions then
          begin
            FConnection.TranRollback(E);
          end;
          
          raise;
        end;
      end;
    finally
      FreeAndNil(LScriptParser);
    end;
  finally
    FreeAndNil(LStatementWorker);
  end;
end;

procedure TibeSQLScript.TExecWorker.Execute(const AStatement: PibeSQLStatement);
begin
  FQuery.SQL := AStatement.SQLStatement;
  FQuery.Execute();
end;

// ======================= [ TibeSQLStatementWorker ] ==========================

constructor TibeSQLStatementWorker.Create(AConnection: TibeDALConnection); 
begin
  inherited Create();

  FQuery := AConnection.DAOQueryCreate();
end;

destructor TibeSQLStatementWorker.Destroy(); 
begin
  inherited Destroy();

  FreeAndNil(FQuery);
end;

// ======================== [ TibeSQLScriptParser ] ============================

procedure TibeSQLScriptParser.Execute(const AFileName: TFileName; AWorker: TibeSQLStatementWorker);
var
  LStream: TFileStream;
  LReader: TStreamReader;
  LStringlist: TStringList;
  LStatement: TibeSQLStatement;
  LRow, LTrimmedRow, LTimestamp, LSQLText: string;
  LCurrentRowNum: Integer;
  LEndOfStatement: Boolean;
  LLogFileName, LScriptFileName: TFileName;
  LLogStream: TFileStream;
  LLogWriter: TStreamWriter;
  LSeconds: Extended;
begin
  LCurrentRowNum := 0;
  LStatement.RowNum := 1;
  LStatement.Index := 0;
  LStatement.ScriptFileName := AFileName;
  LLogFileName := ExtractFileJustName(AFileName);
  LLogFileName := AddBackslash(ExtractFilePath(AFileName)) + LLogFileName + '-parsed' + ExtractFileExt(AFileName);

  LLogStream := TFileStream.Create(LLogFileName, fmOpenWrite + fmCreate);
  try
    LLogWriter := TStreamWriter.Create(LLogStream);
    try
      LStringlist := TStringList.Create();
      try
        LStream := TFileStream.Create(AFileName, fmOpenRead);
        try
          LReader := TStreamReader.Create(LStream);
          try
            LStringlist.Clear();
            LEndOfStatement := False;
            while not LReader.EndOfStream do
            begin
              Inc(LCurrentRowNum);
          
              LRow := LReader.ReadLine();
              LTrimmedRow := Trim(LRow);
              if LTrimmedRow <> '' then
              begin
                LEndOfStatement := LTrimmedRow[Length(LTrimmedRow)] = ';';
          
                if LEndOfStatement then
                begin
                  if LStringlist.Count = 0 then
                  begin
                    // neues Statement: Zeilennummer merken
                    LStatement.RowNum := LCurrentRowNum;
                  end;
                  // die Zeile markiert das Ende vom Statenemt, enthlt aber mglicherweise 
                  // noch das Ende vom Statement. Die Zeile an das Statement anhngen
                  LStringlist.Add(LRow);
                end else
                begin
                  LEndOfStatement := SameText(LTrimmedRow, 'go');
                  if not LEndOfStatement then
                  begin
                    if LStringlist.Count = 0 then
                    begin
                      // neues Statement: Zeilennummer merken
                      LStatement.RowNum := LCurrentRowNum;
                    end;
                    LStringlist.Add(LRow);
                  end;
                end;
              end;

              if (LEndOfStatement or LReader.EndOfStream) and (LStringlist.Count > 0) then
              begin
                LSQLText := TrimRight(LStringlist.Text);
                // prfen, ob ein externes Scriptfile nachgeladen werden soll
                if SameText(Copy(LSQLText, 1, Length(GC_INCLUDE_PREFIX)), GC_INCLUDE_PREFIX) then
                begin
                  LLogWriter.WriteLine(LSQLText);
                  // den Dateinamen vom Script extrahieren
                  LScriptFileName := Copy(LSQLText, Length(GC_INCLUDE_PREFIX) + 1, Length(LSQLText) - Length(GC_INCLUDE_PREFIX) - 1);
                  // mit dem Pfad aus dem bergebenen Script den Dateinamen vervollstndigen
                  LScriptFileName := AddBackslash(ExtractFilePath(AFileName)) + LScriptFileName;
                  // Das Scriptfile einlesen
                  LStringlist.LoadFromFile(LScriptFileName);
                  // Den Inhalt vom Scriptfile als SQL Statement verwenden
                  LSQLText := LStringlist.Text;
                end;

                LStatement.SQLStatement := LSQLText;
                LStringlist.Clear();
                LTimestamp := FormatDateTime(ShortDateFormat + ' hh:mm:ss.zzz', Now());
                LLogWriter.WriteLine('-- %s Statement #%d, start at Row #%d', [LTimestamp, LStatement.Index + 1, LStatement.RowNum]);
                LLogWriter.WriteLine(LStatement.SQLStatement);

                LSeconds := QueryElapsedSeconds();

                try
                  try            
                    AWorker.Execute(@LStatement); // das Statement ausfhren
                  except
                    on E: Exception do
                    begin
                      LLogWriter.WriteLine(E.ClassName + ': ' + E.Message);
                      raise;
                    end;
                  end;
                finally
                  LSeconds := QueryElapsedSeconds(LSeconds);
                  LTimestamp := FormatDateTime(ShortDateFormat + ' hh:mm:ss.zzz', Now());
                  LLogWriter.WriteLine(Format('-- %s %f sec elapsed', [LTimestamp, LSeconds]) + #13#10);
                end;

                LEndOfStatement := False;
                Inc(LStatement.Index);
              end;
            end;
          finally
            FreeAndNil(LReader);
          end;
        finally
          FreeAndNil(LStream);
        end;
      finally
        FreeAndNil(LStringlist);
      end;
    finally
      FreeAndNil(LLogWriter);
    end;
  finally
    FreeAndNil(LLogStream);
  end;
end;

// ================ [ TibeSQLStatementSaveToFileWorker ] =======================

destructor TibeSQLStatementSaveToFileWorker.Destroy();
begin
  inherited Destroy();

//  FWriter.Flush();

  FreeAndNil(FWriter);
  FreeAndNil(FStream);
end;

procedure TibeSQLStatementSaveToFileWorker.Execute(const AStatement: PibeSQLStatement);
//var
//  LFileName: TFileName;
begin
//  if not Assigned(FStream) then
//  begin
//    // Stream und Writer sind noch nicht angelegt
//    LFileName := ExtractFileJustName(AStatement.ScriptFileName);
//    LFileName := AddBackslash(ExtractFilePath(AStatement.ScriptFileName)) + LFileName + '-parsed' + ExtractFileExt(AStatement.ScriptFileName);
//
//    FStream := TFileStream.Create(LFileName, fmOpenWrite + fmCreate);
//    FWriter := TStreamWriter.Create(FStream);
//  end;
//
//  FWriter.WriteLine('-- Statement #%d, start at Row #%d', [AStatement.Index + 1, AStatement.RowNum]);
//  FWriter.WriteLine(AStatement.SQLStatement);
end;

end.
