// ============================================================================
//
// Unit:    ibeVariant
// Inhalt:  Formular
// Aufgabe: Funktionsbibliothek
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
{ Copyright 2005 bis 2013 by Klaus Etscheid }
unit ibeVariant;

interface

Uses
	SysUtils, StrUtils, Variants, DB, ADODB,
	ibeTypes, ibeString;

{ **************************************************************************** }
{ *                                                                          * }
{ * Methoden fr den Umgang mit Varianten                                    * }
{ *                                                                          * }
{ **************************************************************************** }

Function VarHasValue( v: Variant ): Boolean;
Function VarsAreEqual( v1, v2: Variant ): Boolean;
Function CompareVariants( v1, v2: Variant ): Integer;

Function VarAsBoolean( v: Variant; ADefault: Boolean = False ): Boolean;
Function VarAsInteger( v: Variant; ADefault: Integer = 0 ): Integer;
Function VarAsDouble( v: Variant; ADefault: Double = 0.0 ): Double;
Function VarAsDate( v: Variant; ADefault: TDateTime = 0.0 ): TDateTime;
Function VarAsString( v: Variant; Const ADefault: WideString = '' ): string;
function VarAsTypedVariant(v: Variant; AType: TibeDatatype): Variant;

Function StringToTypedVariant( Const AValue: WideString; DataType: TibeDatatype ): Variant;

{ **************************************************************************** }
{ *                                                                          * }
{ * Varianten in XML-konformen Stringreprsentationen abbilden.              * }
{ *                                                                          * }
{ **************************************************************************** }

Function VarAsXMLString( v: Variant ): ibeWideString;
Function XMLStringAsVariant( Const s: ibeWideString; DataType: TibeDatatype ): Variant;

{ **************************************************************************** }
{ *                                                                          * }
{ * Methoden fr den Umgang mit Varianten und Datenbankobjekten              * }
{ *                                                                          * }
{ **************************************************************************** }

Function IfThen( AValue: Boolean; ATrue, AFalse: Variant ): Variant; Overload;
Procedure VarToDBParam( v: Variant; p: TParam ); Overload;
Procedure VarToDBParam( v: Variant; p: TParameter ); Overload;
Procedure VarToDBField( v: Variant; f: TField );

Function VarTypeToIbeDatatype( VarType: TVarType ):	TibeDatatype;
Function ibeDatatypeFromVariant( v: Variant ):	TibeDatatype;

implementation

Function VarHasValue( v: Variant ): Boolean;
Begin
	Result := Not ((VarType( v ) = varNull) Or (VarType( v ) = varEmpty));
End;

Function VarsAreEqual( v1, v2: Variant ): Boolean;
Var i, min1, min2, max1, max2: Integer;
Begin
	If VarIsArray( v1 ) Or VarIsArray( v2 ) Then Begin
		If VarIsArray( v1 ) And VarIsArray( v2 ) Then Begin
			min1 := VarArrayLowBound( v1, 1 );
			max1 := VarArrayHighBound( v1, 1 );
			min2 := VarArrayLowBound( v2, 1 );
			max2 := VarArrayHighBound( v2, 1 );
			Result := (min1 = min2) And (max1 = max2);
			If Result Then Begin
				For i := min1 To max1 Do Begin
					Result := VarsAreEqual( v1[i], v2[i] );
					If Not Result Then Break; 
				End;
      End;
		End Else Begin
			Result := False;
		End;
	End Else Begin
		Result := v1 = v2;
	End;
End;

Function CompareVariants( v1, v2: Variant ): Integer;
Var vr: TVariantRelationship;
Begin
	vr	:= VarCompareValue( v1, v2 );
	Case vr Of
		vrEqual:				Result :=  0;
		vrLessThan:			Result := -1;
		vrGreaterThan:	Result :=  1;
		vrNotEqual:			Result := IfThen( VarHasValue( v1 ), 1, -1 );	// Tritt nur auf, wenn
															// entweder v1 oder v2 nicht zugewiesen wurden. In diesem Fall
															// ist der zugewiesene Wert grer als der nicht zugewiesene
	Else							Result := 0;	// Alles andere: Variablen sind gleich
	End;
End;

Function VarAsBoolean( v: Variant; ADefault: Boolean ): Boolean;
Begin
	If VarHasValue( v ) Then Result := VarAsType( v, varBoolean )
	Else                     Result := ADefault;
End;

Function VarAsInteger( v: Variant; ADefault: Integer ): Integer;
Begin
	If VarHasValue( v ) Then Result := v
	Else                     Result := ADefault;
End;

Function VarAsDouble( v: Variant; ADefault: Double ): Double;
Begin
	If VarHasValue( v ) Then Result := v
	Else                     Result := ADefault;
End;

Function VarAsDate( v: Variant; ADefault: TDateTime ): TDateTime;
Begin
	If VarHasValue( v ) Then Result := VarAsType( v, varDate )
	Else										 Result := ADefault
End;

Function VarAsString( v: Variant; Const ADefault: WideString ): string;
Begin
	If VarHasValue( v ) Then Result := VarToStr( v )
	Else                		 Result := ADefault;
End;

function VarAsTypedVariant(v: Variant; AType: TibeDatatype): Variant;
begin
	Case AType Of
		dtString:		Result := VarAsString(v);
		dtInteger:	Result := VarAsInteger(v);
		dtBoolean:	Result := VarAsBoolean(v);
		dtFloat:		Result := VarAsDouble(v);
		dtDateTime:	Result := VarAsDate(v);
	Else					Result := v;
	End;
end;

Function StringToTypedVariant( Const AValue: WideString; DataType: TibeDatatype ): Variant;
Begin
	Case DataType Of
		dtString:		Result := AValue;
		dtInteger:	Result := StrToInt( AValue );
		dtBoolean:	Result := StrToBool( AValue );
		dtFloat:		Result := StrToFloat( AValue );
		dtDateTime:	Result := StrToDateTime( AValue );
	Else					Result := AValue;
	End;
End;

Function VarAsXMLString( v: Variant ): ibeWideString;
var ibeType: TibeDatatype;
Begin
	ibeType := VarTypeToIbeDatatype(VarType(v));
	Case ibeType Of
		dtDateTime:	Result := nDateTimeToStr( v );
		dtFloat:		Result :=	nFloatToStr( v );
		dtBoolean:	Result := nBooleanToStr( v );
	Else					Result := VarToWideStr( v );
	End;
End;

Function XMLStringAsVariant( Const s: ibeWideString; DataType: TibeDatatype ): Variant;
Begin
	If Trim(s) = '' Then Begin
		Result := Null;
	End Else Begin
		Case DataType Of
			dtDateTime:		Result := nStrToDateTime( s );
			dtFloat:			Result :=	nStrToFloat ( s );
			dtBoolean:		Result := nStrToBoolean( s );
			dtInteger:		Result := StrToInt( s );
		Else						Result := s;
		End;
	End;
End;

Function IfThen( AValue: Boolean; ATrue, AFalse: Variant ): Variant;
Begin
	If AValue Then Result := ATrue
	Else					 Result := AFalse;
End;

Procedure VarToDBParam( v: Variant; p: TParam );
Begin
	If VarHasValue( v ) Then p.Value := v
	Else                     p.Clear;
End;

Procedure VarToDBParam( v: Variant; p: TParameter );
Begin
	p.Value := v;
End;

Procedure VarToDBField( v: Variant; f: TField );
Begin
	If VarHasValue( v ) Then f.Value := v
	Else                     f.Clear;
End;

Function VarTypeToIbeDatatype( VarType: TVarType ):	TibeDatatype;
Begin
	Case VarType Of
	{ Allbemeine Ganzzahl-Typen }
		varSmallint,													// 16-Bit-Integer mit Vorzeichen (Typ Smallint in Delphi, short in C++).
		varInteger,														// 32-Bit-Integer mit Vorzeichen (Typ Integer in Delphi, int in C++).
		varSingle,														// Gleitkommawert mit einfacher Genauigkeit (Typ Single in Delphi, float in C++).
		varShortInt,													// 8-Bit-Integer mit Vorzeichen (Typ ShortInt in Delphi, char mit Vorzeichen in C++).
		varByte,															// Byte
		varWord,															// 16-Bit-Wert ohne Vorzeichen (Word)
		varLongWord,													// 32-Bit-Integer ohne Vorzeichen (Typ LongWord in Delphi, long ohne Vorzeichen in C++).
		varInt64:			Result	:= dtInteger;		// 64-Bit-Integer mit Vorzeichen (Typ Int64 in Delphi, __int64 in C++).
	{ Allgemeine Gleitkommawerte }
		varDouble,														// Gleitkommawert doppelter Genauigkeit (double)
		varCurrency:	Result := dtFloat;			// Whrungs-Gleitkommawert (Typ Currency).
	{ Allgemeine Datum-/Zeit-Typen }
		varDate:			Result := dtDateTime;		// Datums-/Zeitwert (Typ TDateTime).
	{ Allgemeinte Zeichenketten }
		varString,														// Referenz auf dynamischen String (nicht COM-kompatibel)
    varUString,                           // Unicode string
		varOleStr,														// Referenz auf dynamisch zugeordneten UNICODE-String.
		varStrArg:		Result := dtString;			// COM-kompatibler String.
	{ Allgemeine Boolesche Typen }
		varBoolean:		Result := dtBoolean;		// Boolescher 16-Bit-Wert (Typ WordBool).
	{ Von TibeDatatype nicht untersttzte Typen }
		varEmpty,															// Die Variante hat den Wert Unassigned.
		varNull,															// Die Variante hat den Wert Null.
		varVariant,														// Variante.
		varUnknown,														// Referenz auf unbekanntes Objekt (IInterface- oder IUnknown-Schnittstellenzeiger).
		varDispatch,													// Referenz auf Automationsobjekt (IDispatch-Schnittstellenzeiger).
		varError,															// Betriebssystem-Fehlercode.
		varAny:				Result := dtNone;				// Beliebiger CORBA-Wert.
	Else						Result := dtNone;
	End;
End;

Function ibeDatatypeFromVariant( v: Variant ):	TibeDatatype;
Begin
	Result	:= VarTypeToIbeDatatype( VarType( v ) );
End;

end.
