// ============================================================================
//
// Unit:    TestDataAccessObject
// Inhalt:  Formular
// Aufgabe: Abstrakter DUnit Test fr den Data Abstraction Layer. Fr die
//          konkreten DAL-Klassen mssen Unit-Tests als Ableitung dieser 
//          Testklasse erstellt werden
//
//   (c) by Klaus Etscheid since 2015
// ============================================================================
unit TestDataAccessObject;
{

  Delphi DUnit Testfall
  ----------------------
  Diese Unit enthlt ein Codegerst einer Testfallklasse, das vom Testfall-Experten
  erzeugt wurde. ndern Sie den erzeugten Code, damit die Methoden aus der 
  getesteten Unit korrekt eingerichtet und aufgerufen werden.

}

interface

uses
  TestFramework, SysUtils, Variants, Classes,
  ibeString, ibeFileUtil, ibeLogfile,
  ibeDALConnection, ibeDALConnectionManager, ibeTypes;

type
  // Testmethoden fr Klasse TibeDataAccessObject
  
  TestTibeDataAccessObject = class(TTestCase)
  strict private
  	FConnectionManager: TibeDALConnectionManager;
    FConnection: TibeDALConnection;
    FLogger: TibeLogFile;
    FBinPath, FProgPath, FCfgPath: TFileName;
    FIBEConnectionName, FDBName: string;
    FInitScriptDDLWithTransaction: Boolean;
    FInitScriptDDL, FInitScriptDML: TFileName;
    procedure RetrieveIdentityInsertMethod();
    procedure RetrieveSequenceMethod();
    procedure TempTableHandlingMethod();
  protected
    function CreateAndOpenConnection(const ADBName: string): TibeDALConnection;
    function InternalTestDAOTempTableHandling(AConnection: TibeDALConnection): Boolean; virtual; 
    property IBEConnectionName: string read FIBEConnectionName write FIBEConnectionName;
    property DBName: string read FDBName write FDBName;
    property InitScriptDDL: TFileName read FInitScriptDDL write FInitScriptDDL;
    procedure DoBeforeInitScriptDDL(AConnection: TibeDALConnection); virtual;
    property InitScriptDML: TFileName read FInitScriptDML write FInitScriptDML;
    property InitScriptDDLWithTransaction: Boolean read FInitScriptDDLWithTransaction write FInitScriptDDLWithTransaction;
  public
    constructor Create(AMethodName: string); override;
    destructor Destroy(); override;
  	property ConnectionManager: TibeDALConnectionManager read FConnectionManager;
    property BinPath: TFileName read FBinPath;
    property ProgPath: TFileName read FProgPath;
    property CfgPath: TFileName read FCfgPath;
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestDAOBasics;
    procedure TestDAOConnectionMethods;     
    procedure TestDAOSimpleScript;          
    procedure TestDAOTransactions;          
    procedure TestDAORetrieveIdentity;
    procedure TestDAORetrieveSequence;      
    procedure TestDAOTempTableHandling;
    procedure TestDAORetrieveTableList;
  end;

implementation

uses
  ibeAppServiceProviderDB,
  ibeDALQuery, ibeSQLScript;

constructor TestTibeDataAccessObject.Create(AMethodName: string); 
var
  LLogFile: TFileName;
begin
  inherited Create(AMethodName);
  FBinPath	:= ExtractFilePath(ParamStr(0));
	FProgPath	:= RemoveLastWord(BinPath, ['\']);
	FCfgPath  := AddBackslash(ProgPath) + 'Config';

  InitScriptDDLWithTransaction := False;

  LLogFile  := AddBackslash(ProgPath) + 'Data\Log';
  CreatePath(LLogFile);
  LLogFile := AddBackslash(LLogFile) + ExtractFileJustName(ParamStr(0)) + '.log';
  FLogger := TibeLogFile.Create(LLogFile);
end;

destructor TestTibeDataAccessObject.Destroy();
begin
  inherited Destroy();

  FreeAndNil(FLogger);
end;

procedure TestTibeDataAccessObject.SetUp;
begin
	FConnectionManager := TibeDALConnectionManager.Create;
  ConnectionManager.CfgFileName := AddBackslash(CfgPath) + 'AppConfig.ini';
end;

procedure TestTibeDataAccessObject.TearDown; 
begin
  FreeAndNil(FConnectionManager);
end;

procedure TestTibeDataAccessObject.TestDAOBasics;
var
  LConnection: TibeDALConnection;
  LQuery: TibeDALQuery;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LQuery := LConnection.DAOQueryCreate();
    try
      CheckNotNull(LQuery, 'Query');
    finally
      FreeAndNil(LQuery);
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

function TestTibeDataAccessObject.CreateAndOpenConnection(const ADBName: string): TibeDALConnection;
var
  LConnection: TibeDALConnection;
begin
  LConnection :=  ConnectionManager.RetriveConnection(nil, IBEConnectionName);

  FLogger.Write('ffne Datenbank "' + ADBName + '"');
  try
    LConnection.ActiveDatabase := ADBName;
    LConnection.Open();
  except
    on E: Exception do
    begin
      FLogger.Write(E.ClassName + ': ' + E.Message);
      raise;
    end;
  end;

  Result := LConnection;
end;

procedure TestTibeDataAccessObject.TestDAOConnectionMethods;
var
  LConnection: TibeDALConnection;
  LResult: ibeWideString;
begin
  FLogger.Write('Testmethode "' + FTestName + '"');

  LConnection :=  ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LResult := LConnection.CVToString(Null);
    CheckEquals('null', LResult, 'CVToString(Null) muss Leerstring liefern');

    LResult := LConnection.CVToString(999);
    CheckEquals('999', LResult, 'CVToString(999)');

    LResult := LConnection.CVToString('string');
    CheckEquals('''string''', LResult, 'CVToString(''string'')');
  
    LResult := LConnection.CVBoole(True);
    CheckNotEquals('', LResult, 'CVBoole(True) darf keinen Leerstring liefern');
  
    LResult := LConnection.CVDate(Date());
    CheckNotEquals('', LResult, 'CVDate(Date()) darf keinen Leerstring liefern');

    LResult := LConnection.CVDateTime(Now());
    CheckNotEquals('', LResult, 'CVDateTime(True) darf keinen Leerstring liefern');

    LResult := LConnection.CVFloat(3.14);
    CheckEquals('3.14', LResult, 'CVFloat(999)');
  finally
    FreeAndNil(LConnection);
  end;
end;

procedure TestTibeDataAccessObject.TestDAOSimpleScript;
var
  LConnection: TibeDALConnection;
  LScript: TibeSQLScript;
  LQuery: TibeDALQuery;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    LConnection.KeepConnection := True;

    LScript := TibeSQLScript.Create(LConnection);
    try
      CheckTrue(True);

      DoBeforeInitScriptDDL(LConnection);

      LConnection.Open();

      LScript.ActivateTransactions := InitScriptDDLWithTransaction;
      LScript.FileName := InitScriptDDL;
      LScript.Execute();

      LConnection.ActiveDatabase := DBName;

      LScript.ActivateTransactions := True;    
      LScript.FileName := InitScriptDML;
      LScript.Execute();
    finally
      FreeAndNil(LScript);
    end;

    CheckNotNull(LConnection, 'Connection');

    LQuery := LConnection.DAOQueryCreate();
    try
      CheckNotNull(LQuery, 'Query');
      
      LQuery.SQL := 'select * from testBasics where ID = ''00010''';
      LQuery.Open();

      CheckEquals(1, LQuery.DataSet.RecordCount, 'Anzahl Datenstze');
      CheckEquals('Klaus', LQuery.DataSet.FieldByName('Name').AsString, 'Wert von Feld Name');
    finally
      FreeAndNil(LQuery);
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

procedure TestTibeDataAccessObject.DoBeforeInitScriptDDL(AConnection: TibeDALConnection); 
begin
end;

procedure TestTibeDataAccessObject.TestDAOTransactions;
var
  LConnection: TibeDALConnection;
  LQuery: TibeDALQuery;
  LCtrlTran: Boolean;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LConnection.ActiveDatabase := DBName;
    LConnection.Open();

    LQuery := LConnection.DAOQueryCreate();
    try
      CheckNotNull(LQuery, 'Query');
      
      // 1. Test: Datensatz anlegen. Nach Rollback darf der Datensatz nicht mehr vorhanden sein
      LCtrlTran := LConnection.TranStart();
      try
        CheckTrue(LCtrlTran, 'Transaktionssteuerung aktiv');

        LQuery.SQL := 'insert into testBasics (ID, Name) values (''TRAN'', ''Klaus'')';
        LQuery.Execute();

        LConnection.TranRollback();
      
        LQuery.SQL := 'select count(*) as Anz from testBasics where ID = ''TRAN''';
        LQuery.Open();
        CheckEquals(LQuery.DataSet.FieldByName('Anz').AsInteger, 0, 'Anzahl Datenstze nach rollback');
      except
        on E: Exception do
        begin
          LConnection.TranRollback(E);
          raise;
        end;
      end;

      // 2. Test: Datensatz anlegen. Nach Commit muss der Datensatz noch da sein
      LCtrlTran := LConnection.TranStart();
      try
        CheckTrue(LCtrlTran, 'Transaktionssteuerung aktiv');

        LQuery.SQL := 'insert into testBasics (ID, Name) values (''TRAN'', ''Klaus'')';
        LQuery.Execute();

        LConnection.TranCommit();
      
        LQuery.SQL := 'select count(*) as Anz from testBasics where ID = ''TRAN''';
        LQuery.Open();
        CheckEquals(LQuery.DataSet.FieldByName('Anz').AsInteger, 1, 'Anzahl Datenstze nach commit');
      except
        on E: Exception do
        begin
          LConnection.TranRollback(E);
          raise;
        end;
      end;

      // 3. Test: Datensatz von oben lschen. Nach Rollback muss der Datensatz noch da sein
      LCtrlTran := LConnection.TranStart();
      try
        CheckTrue(LCtrlTran, 'Transaktionssteuerung aktiv');

        LQuery.SQL := 'delete from testBasics where ID = ''TRAN''';
        LQuery.Execute();

        LConnection.TranRollback();
      
        LQuery.SQL := 'select count(*) as Anz from testBasics where ID = ''TRAN''';
        LQuery.Open();
        CheckEquals(LQuery.DataSet.FieldByName('Anz').AsInteger, 1, 'Anzahl Datenstze nach rollback');
      except
        on E: Exception do
        begin
          LConnection.TranRollback(E);
          raise;
        end;
      end;

      // 4. Test: Datensatz von oben lschen. Nach Commit muss der Datensatz weg sein
      LCtrlTran := LConnection.TranStart();
      try
        CheckTrue(LCtrlTran, 'Transaktionssteuerung aktiv');

        LQuery.SQL := 'delete from testBasics where ID = ''TRAN''';
        LQuery.Execute();

        LConnection.TranCommit();
      
        LQuery.SQL := 'select count(*) as Anz from testBasics where ID = ''TRAN''';
        LQuery.Open();
        CheckEquals(LQuery.DataSet.FieldByName('Anz').AsInteger, 0, 'Anzahl Datenstze nach rollback');
      except
        on E: Exception do
        begin
          LConnection.TranRollback(E);
          raise;
        end;
      end;
    finally
      FreeAndNil(LQuery);
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

procedure TestTibeDataAccessObject.TestDAORetrieveIdentity;
var
  LConnection: TibeDALConnection;
  LQuery: TibeDALQuery;
  LID1, LID2: Integer;
  LCtrlTran: Boolean;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LConnection.ActiveDatabase := DBName;
    LConnection.Open();

    if LConnection.IsSupportedIentityAfterInsert then
    begin
      LQuery := LConnection.DAOQueryCreate();
      try
        CheckNotNull(LQuery, 'Query');

        LCtrlTran := LConnection.TranStart();
        try
          LQuery.SQL := 'insert into testIdentity (Name) values (''Klaus'')';
          LQuery.Execute();
          LID1 := LConnection.RetrieveIdentityValueAfterInsert('testIdentity_PK_Sequence');
      
          LQuery.SQL := 'insert into testIdentity (Name) values (''Ute'')';
          LQuery.Execute();
          LID2 := LConnection.RetrieveIdentityValueAfterInsert('testIdentity_PK_Sequence');

          CheckTrue(LID1 <> LID2, 'IdentityInsert, Wert ID1=' + IntToStr(LID1) + ' ungleich ID2=' + IntToStr(LID2));

          LQuery.SQL := 'select * from testIdentity where ID = ' + IntToStr(LID2);
          LQuery.Open();

          CheckEquals(1, LQuery.DataSet.RecordCount, 'Anzahl Datenstze');
          CheckEquals('Ute', LQuery.DataSet.FieldByName('Name').AsString, 'Wert von Feld Name');

          LConnection.TranCommit();
        except
          on E: Exception do
          begin
            LConnection.TranRollback(E);
            raise;
          end;
        end;
      finally
        FreeAndNil(LQuery);
      end;
    end else
    begin
      FConnection := LConnection;
      try
        CheckException(RetrieveIdentityInsertMethod, EibeDALConnectionNotSupported);
      finally
        FConnection := nil;
      end;
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

procedure TestTibeDataAccessObject.RetrieveIdentityInsertMethod();
begin
  FConnection.RetrieveIdentityValueAfterInsert();
end;

procedure TestTibeDataAccessObject.TestDAORetrieveSequence;
const
  LSequenceName = 'testSequence';
var
  LConnection: TibeDALConnection;
  LVal1, LVal2: Integer;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LConnection.ActiveDatabase := DBName;
    LConnection.Open();

    if LConnection.IsSupportedSequence then
    begin
      LVal1 := LConnection.RetrieveSequenceValue(LSequenceName);
      LVal2 := LConnection.RetrieveSequenceValue(LSequenceName);

      CheckTrue(LVal1 < LVal2, 'Sequence: zweiter Wert muss grer als der erste sein');
    end else
    begin
      FConnection := LConnection;
      try
        CheckException(RetrieveSequenceMethod, EibeDALConnectionNotSupported);
      finally
        FConnection := nil;
      end;
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

procedure TestTibeDataAccessObject.RetrieveSequenceMethod();
begin
  FConnection.RetrieveSequenceValue('Meine_Sequence');
end;

procedure TestTibeDataAccessObject.TestDAOTempTableHandling;
var
  LConnection: TibeDALConnection;
  LCtrlTran: Boolean;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LConnection.ActiveDatabase := DBName;
    LConnection.Open();

    if LConnection.IsSupportedTempTable then
    begin
        LCtrlTran := LConnection.TranStart();
        try
          CheckTrue(InternalTestDAOTempTableHandling(LConnection), 'TempTableHandling');

          LConnection.TranCommit();
        except
          on E: Exception do
          begin
            LConnection.TranRollback(E);
            raise;
          end;
        end;
    end else
    begin
      FConnection := LConnection;
      try
        CheckException(TempTableHandlingMethod, EibeDALConnectionNotSupported);
      finally
        FConnection := nil;
      end;
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

function TestTibeDataAccessObject.InternalTestDAOTempTableHandling(AConnection: TibeDALConnection): Boolean;
begin
  Result := False;
end;

procedure TestTibeDataAccessObject.TempTableHandlingMethod();
begin
  FConnection.RetrieveTempTableExists('testTemp');
end;

procedure TestTibeDataAccessObject.TestDAORetrieveTableList;
var
  LConnection: TibeDALConnection;
  LQuery: TibeDALQuery;
  LStringList: TStringList;
  LCtrlTran: Boolean;
begin
  LConnection := ConnectionManager.RetriveConnection(nil, IBEConnectionName);
  try
    CheckNotNull(LConnection, 'Connection');

    LConnection.ActiveDatabase := DBName;
    LConnection.Open();

    LStringList := TStringList.Create();
    try

      LQuery := LConnection.DAOQueryCreate();
      try
        LCtrlTran := LConnection.TranStart();
        try
          CheckNotNull(LQuery, 'Query');

          LConnection.RetrieveTableList(LStringList);
    
          CheckNotEquals(0, LStringList.Count, 'RetrieveTableList(), Anzahl > 0');
          CheckNotEquals(-1, LStringList.IndexOf('testBasics'), 'Index fr Tabelle testBasics <> -1');

          LConnection.TranCommit();
        except
          on E: Exception do
          begin
            LConnection.TranRollback(E);
            raise;
          end;
        end;
      finally
        FreeAndNil(LQuery);
      end;
    finally
      FreeAndNil(LStringList);
    end;
  finally
    FreeAndNil(LConnection);
  end;
end;

end.
